import requests
import zipfile
import os
import pandas as pd
from io import BytesIO


class Lahman(object):
    """
    Pull data from [Sean Lahman's database](http://www.seanlahman.com/baseball-archive/statistics/), also hosted by
    Chadwick Bureau on GitHub -- our new source):
    """

    _handle = None
    _url = "https://github.com/chadwickbureau/baseballdatabank/archive/master.zip"
    _base_string = os.path.join("baseballdatabank-master", "core")

    def __init__(self):
        self.z = self.__get_lahman_zip()

    def __get_lahman_zip(self):
        """
        Retrieve the Lahman database zip file, returns None if file already exists in cwd.
        If we already have the zip file, keep re-using that.

        :return: zipFile
        """
        if os.path.exists(self._base_string):
            self._handle = None
        elif not self._handle:
            s = requests.get(self._url, stream=True)
            self._handle = zipfile.ZipFile(BytesIO(s.content))
        return self._handle

    def __download_lahman(self):
        """
        Download entire lahman db to present working directory

        :return: None
        """
        z = self.__get_lahman_zip()
        if z is not None:
            z.extractall(self._handle)
            z = self.__get_lahman_zip()

    @property
    def parks(self):
        """
        List of major league ballparks

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "Parks.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def all_star_full(self):
        """
        Get All Star Appearances

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "AllstarFull.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def appearances(self):
        """
        Details on the positions a player appeared at

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "Appearances.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def awards_managers(self):
        """
        Awards won by managers

        :return: DataFrame
        """
        self.z = self.get_lahman_self.zip()
        f = os.path.join(self._base_string, "AwardsManagers.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def awards_players(self):
        """
        awards won by players

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "AwardsPlayers.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def awards_share_managers(self):
        """
        award voting for manager awards

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "AwardsShareManagers.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def awards_share_players(self):
        """
        award voting for player awards

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "AwardsSharePlayers.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def batting(self):
        """
        batting statistics

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "Batting.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def batting_post(self):
        """
        post-season batting statistics

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "BattingPost.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def college_playing(self):
        """
        list of players and the colleges they attended

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "CollegePlaying.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def fielding(self):
        """
        fielding statistics

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "Fielding.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def fielding_of(self):
        """
        Outfield position data

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "FieldingOF.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def fielding_of_split(self):
        """
        post-season fielding data

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "FieldingOFsplit.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def fielding_post(self):
        """
        post-season fielding data

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "FieldingPost.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def hall_of_fame(self):
        """
        Hall of Fame voting data

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "HallOfFame.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def home_games(self):
        """
        Number of homegames played by each team in each ballpark

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "HomeGames.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def managers(self):
        """
        Managerial statistics

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "Managers.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def managers_half(self):
        """
        Split season data for managers

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "ManagersHalf.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    # Alias for people -- the new name for master
    @property
    def master(self):
        """
        Player names, DOB, and biographical info

        :return: DataFrame
        """
        return self.people()

    @property
    def people(self):
        """
        Player names, DOB, and biographical info

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "People.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def pitching(self):
        """
        Pitching statistics

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "Pitching.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def pitching_post(self):
        """
        Post-season pitching statistics

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "PitchingPost.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def salaries(self):
        """
        post-season pitching statistics

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "Salaries.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def schools(self):
        """
        list of colleges that players attended

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "Schools.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',',
                           quotechar='"')  # different here bc of doublequotes used in some school names
        return data

    @property
    def series_post(self):
        """
        post-season series information

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "SeriesPost.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def teams(self):
        """
        yearly player_type and standings

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "Teams.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def teams_franchises(self):
        """
        Team Franchise

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "TeamsFranchises.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data

    @property
    def teams_half(self):
        """
        franchise information

        :return: DataFrame
        """
        f = os.path.join(self._base_string, "TeamsHalf.csv")
        data = pd.read_csv(f if self.z is None else self.z.open(f), header=0, sep=',', quotechar="'")
        return data



from plyball.fangraph import FanGraphs
from plyball.ottoneu import Ottoneu
import pandas as pd

import unittest


class TestFanGraphs(unittest.TestCase):

    def test_pitching_table(self):
        fg = FanGraphs()
        pt = fg.get_pitching_table(2021)

        self.assertEqual(type(pt), pd.DataFrame)
        self.assertEqual(pt.shape, (885, 324))


class TestOttoneu(unittest.TestCase):

    def test_pitching_table(self):
        ot = Ottoneu(186)
        players = ot.players()

        self.assertEqual(type(players['batterResults']), pd.DataFrame)
        self.assertEqual(players['info'].shape[1], 21)
        self.assertEqual(players['stat'].shape[1], 18)


if __name__ == '__main__':
    unittest.main()



import pandas as pd


def get_player_map():
    sheet_id = '1JgczhD5VDQ1EiXqVG-blttZcVwbZd5_Ne_mefUGwJnk'
    sheet_name = 'PLAYERIDMAP'
    url = f'https://docs.google.com/spreadsheets/d/{sheet_id}/gviz/tq?tqx=out:csv&sheet={sheet_name}'
    return pd.read_csv(url)




import logging
import re
from typing import Literal

import numpy as np
import pandas as pd
import requests
from bs4 import BeautifulSoup


class FanGraphs(object):
    """
    Class for scraping data from FanGraphs

    """
    _urls = {
            'leaders':     'https://www.fangraphs.com/leaders.aspx?{}',
            'milb_stats':  'https://www.fangraphs.com/api/leaders/minor-league/data?{}',
            'projections': 'https://www.fangraphs.com/api/projections?{}'
    }

    logger = logging.getLogger('fangraphs')

    def __init__(self):
        pass

    def __get_leaders_html(self,
                           player_type: Literal['bat', 'pit'],
                           start_season: int,
                           end_season: int = None,
                           league: int = 'all',
                           qualified: int = 'y',
                           split_season: int = 1,
                           **kwargs):
        """
        Get Leaderboard HTML form Fangraphs to convert into a DataFrame

        :param player_type: pit or bat
        :type player_type: Literal['bat','pit']
        :param start_season:
        :param end_season:
        :param league:
        :param qualified: The minimum pla
        :param split_season:
        :return: BeautifulSoup
        """

        if player_type == 'pit':
            columns = ['c'] + [str(num) for num in range(3, 322)]
        elif player_type == 'bat':
            columns = ['c'] + [str(num) for num in range(3, 305)]
        else:
            columns = ['c']

        parameters = {
                'pos':     'all',
                'stats':   player_type,
                'lg':      league,
                'qual':    qualified,
                'type':    ','.join(columns),
                'season':  start_season,
                'month':   0,
                'season1': end_season,
                'ind':     split_season,
                'team':    ','.join(kwargs.get('team', [''])),
                'rost':    kwargs.get('rost', ''),
                'age':     kwargs.get('age', ''),
                'filter':  kwargs.get('filter', ''),
                'players': kwargs.get('players', '0'),
                'page':    '1_999999999'
        }

        self.logger.info(self._urls['leaders'].format('&'.join(['{}={}'.format(k, v) for k, v in parameters.items()])))
        s = requests.get(
                self._urls['leaders'].format('&'.join(['{}={}'.format(k, v) for k, v in parameters.items()]))).content
        self.logger.info(self._urls['leaders'].format('&'.join(['{}={}'.format(k, v) for k, v in parameters.items()])))
        return BeautifulSoup(s, "lxml")

    def __get_leader_table(self,
                           player_type: Literal['pit', 'bat'],
                           start_season: int,
                           **kwargs) -> pd.DataFrame:

        soup = self.__get_leaders_html(player_type, start_season, **kwargs)
        table = soup.find('table', {
                'class': 'rgMasterTable'
        })

        __data = []
        __headings = [row.text.strip() for row in table.find_all('th')[1:]]
        __headings.append('player_id')

        fb_perc_indices = [i for i, j in enumerate(__headings) if j == 'FB%']
        __headings[fb_perc_indices[0]] = 'flyball_%'
        __headings[fb_perc_indices[1]] = 'fastball_%'
        __data.append(__headings)
        table_body = table.find('tbody')
        rows = table_body.find_all('tr')

        for row in rows:
            player_id = 'NA'
            cols = row.find_all('td')
            table_cols = []
            for ele in cols:
                table_cols.append(ele.text.strip())
                for link in ele.find_all('a'):
                    url = link.get('href', None)
                    if 'playerid' in url:
                        player_id = url.split('?')[1].split('&')[0].split('=')[1]
            table_cols.append(player_id)
            __data.append([ele for ele in table_cols[1:]])

        __data = pd.DataFrame(data=__data, columns=__headings)[1:]

        # replace empty strings with NaN
        __data.replace(r'^\s*$', np.nan, regex=True, inplace=True)

        # convert percentage strings to floats   FB% duplicated
        percentages = [column for column in __data.columns if '%' in column]
        for col in percentages:
            if not __data[col].empty:
                if pd.api.types.is_string_dtype(__data[col]):
                    __data[col] = __data[col].str.strip(' %')
                    __data[col] = __data[col].str.strip('%')
                    __data[col] = pd.to_numeric(__data[col], errors='coerce') / 100.

        __data.columns = [
                re.sub(r"\W", '_',
                       column.replace('%', '_percent').replace('-', '_negative_').replace('+', '_positive_').replace(
                               '__',
                               '_'))
                for column in __data.columns]
        __data['player_type'] = player_type
        return __data

    def get_pitching_table(self,
                           start_season: int,
                           end_season: int = None,
                           league: str = 'all',
                           minimum_innings_pitched: int = None,
                           split_season: bool = True):
        """
        Get individual pitching data from LeaderBoard.

        :param start_season:
        :param end_season:
        :param league:
        :param minimum_innings_pitched:
        :param split_season:
        :return:
        """
        if start_season is None:
            raise ValueError(
                    "You need to provide at least one season to collect data for. Try pitching_leaders(season) or "
                    "pitching_leaders(start_season, end_season).")
        if end_season is None:
            end_season = start_season
        return self.__get_leader_table(player_type='pit',
                                       start_season=start_season,
                                       end_season=end_season,
                                       league=league,
                                       qualified='y' if minimum_innings_pitched is None else minimum_innings_pitched,
                                       ind=1 if split_season else 0)

    def get_batting_table(self,
                          start_season,
                          end_season: int = None,
                          league: str = 'all',
                          minimum_plate_appearances: int = None,
                          split_season: bool = True):
        """
        Get Individual Batting Stats from the LeaderBoard.

        :param start_season:
        :param end_season:
        :param league:
        :param minimum_plate_appearances:
        :param split_season:
        :return:
        """
        if start_season is None:
            raise ValueError(
                    "You need to provide at least one season to collect data for. Try pitching_leaders(season) or "
                    "pitching_leaders(start_season, end_season).")
        if end_season is None:
            end_season = start_season
        return self.__get_leader_table(player_type='bat',
                                       start_season=start_season,
                                       end_season=end_season,
                                       league=league,
                                       qualified='y' if minimum_plate_appearances is None else minimum_plate_appearances,
                                       ind=1 if split_season else 0)

    def get_team_batting_table(self,
                               start_season: int,
                               end_season: int = None,
                               league: str = 'all',
                               minimum_plate_appearances: int = None,
                               split_season: bool = True):
        """
        Get Team Batting Stats from the LeaderBoard.

        :param start_season: First Season to return data
        :param end_season: Last Season to return data
        :param league: American (AL) or National (NL) Baseball League
        :param minimum_plate_appearances: Minimum number of Plate Appearances (Not supplied only minimum_innings_pitched players will be returned)
        :param split_season: Return Season Splits for players
        :return:
        """
        if start_season is None:
            raise ValueError(
                    "You need to provide at least one season to collect data for. "
                    "Try pitching_leaders(season) or pitching_leaders(start_season, end_season).")
        if end_season is None:
            end_season = start_season
        return self.__get_leader_table(player_type='bat',
                                       start_season=start_season,
                                       end_season=end_season,
                                       league=league,
                                       qualified='y' if minimum_plate_appearances is None else minimum_plate_appearances,
                                       ind=1 if split_season else 0,
                                       team=['0', 'to'])

    def get_team_pitch_table(self,
                             start_season: int,
                             end_season: int = None,
                             league: str = 'all',
                             minimum_innings_pitched: int = None,
                             split_season: bool = True):
        """
        Get Team Pitching Stats from teh LeaderBoard.


        :param start_season: First Season to return data
        :param end_season: Last Season to return data
        :param league: American (AL) or National (NL) Baseball League
        :param minimum_innings_pitched: Minimum number of Pitching Appearances (Not supplied only minimum_plate_appearances players will be returned)
        :param split_season: Return Season Splits for players

        :return: DataFrame
        """
        if start_season is None:
            raise ValueError(
                    "You need to provide at least one season to collect data for. "
                    "Try pitching_leaders(season) or pitching_leaders(start_season, end_season).")
        if end_season is None:
            end_season = start_season
        return self.__get_leader_table(player_type='pit',
                                       start_season=start_season,
                                       end_season=end_season,
                                       league=league,
                                       qualified='y' if minimum_innings_pitched is None else minimum_innings_pitched,
                                       ind=1 if split_season else 0,
                                       team=['0', 'to'])

    def get_milb_pitching_table(self,
                                start_season: int,
                                end_season: int = None,
                                league: str = 'all',
                                minimum_innings_pitched: int = None,
                                split_season: int = 1):
        """
        Get Minor League Pitching Stats from the Prospects sections

        :param start_season: First Season to return data
        :param end_season: Last Season to return data
        :param league: American (AL) or National (NL) Baseball League
        :param minimum_innings_pitched: Minimum number of Plate Appearances (Not supplied only minimum_innings_pitched players will be returned)
        :param split_season: Return Season Splits for players
        :return: DataFrame
        """
        parameters = {
                'pos':         'all',
                'stat':        'pit',
                'lg':          league,
                'qual':        'y' if minimum_innings_pitched is None else minimum_innings_pitched,
                'player_type': ','.join(['0', ]),
                'season':      start_season,
                'season1':     end_season,
                'ind':         split_season,
                'org':         '',
                'splitTeam':   'false'
        }

        json = requests.get(
                url=self._urls['milb_stats'].format(
                        '&'.join(['{}={}'.format(k, v) for k, v in parameters.items()]))).json()

        df = pd.DataFrame(json)
        df['Name'] = df.Name.apply(lambda x: BeautifulSoup(x, 'lxml').a.contents[0])

        return df

    def get_milb_batting_table(self,
                               start_season: int,
                               end_season: int = None,
                               league: str = 'all',
                               minimum_plate_appearances: int = None,
                               split_season: bool = True) -> pd.DataFrame:
        """
        Get Minor League Pitching from the Prospects sections.

        :param start_season: First Season to return data
        :param end_season: Last Season to return data
        :param league: American (AL) or National (NL) Baseball League
        :param minimum_plate_appearances: Minimum number of Plate Appearances (Not supplied only minimum_innings_pitched players will be returned)
        :param split_season: Return Season Splits for players
        :return: DataFrame
        """
        parameters = {
                'pos':         'all',
                'stats':       'pit',
                'lg':          league,
                'qual':        'y' if minimum_plate_appearances is None else minimum_plate_appearances,
                'player_type': ','.join(['0', ]),
                'season':      start_season,
                'season1':     end_season,
                'ind':         1 if split_season else 0,
                'org':         '',
                'splitTeam':   'false'
        }

        json = requests.get(
                url=self._urls['milb_stats'].format(
                        '&'.join(['{}={}'.format(k, v) for k, v in parameters.items()]))).json()

        df = pd.DataFrame(json)
        df['Name'] = df.Name.apply(lambda x: BeautifulSoup(x, 'lxml').a.contents[0])

        return df

    def get_2023_bat_projections(self) -> pd.DataFrame:
        parameters = {
                'type':    'steamer',
                'stats':   'bat',
                'pos':     'all',
                'team':    '0',
                'players': '0',
                'lg':      'all',
        }

        json = requests.get(
                url=self._urls['projections'].format(
                        '&'.join(['{}={}'.format(k, v) for k, v in parameters.items()]))).json()
        df = pd.DataFrame(json)

        return df

    def get_2023_pitching_projections(self) -> pd.DataFrame:
        parameters = {
                'type':    'steamer',
                'stats':   'pit',
                'pos':     'all',
                'team':    '0',
                'players': '0',
                'lg':      'all',
        }

        json = requests.get(
                url=self._urls['projections'].format(
                        '&'.join(['{}={}'.format(k, v) for k, v in parameters.items()]))).json()
        df = pd.DataFrame(json)

        return df



import io
import logging

import pandas as pd
import requests



def get_lookup_table() -> pd.DataFrame:
    """
    Generate a Dataframe of Players

    :return: DataFrame
    """

    logging.info('Gathering player lookup table. This may take a moment.')
    # Construct the URL to the CSV files in the repository
    repo_url = "https://github.com"
    csv_url = repo_url + "/chadwickbureau/register/tree/master/data"

    # Retrieve the HTML contents of the CSV URL
    response = requests.get(csv_url)
    html = response.content

    # Parse the HTML contents to find the URLs of the CSV files
    from bs4 import BeautifulSoup
    soup = BeautifulSoup(html, 'html.parser')
    csv_links = [a['href'] for a in soup.select('a[href$=".csv"]')]

    # Loop through each CSV file and read it into a pandas DataFrame
    register_data = []
    for link in csv_links:
        logging.info(f'Gathering {link}')
        if 'people' in link:
            csv_url = 'https://raw.githubusercontent.com/chadwickbureau/register/master/data/' + link.split('/')[-1]
            logging.info(f'Gathering {csv_url}')
            df = pd.read_csv(csv_url, dtype={
                    'key_sr_nfl': object,
                    'key_sr_nba': object,
                    'key_sr_nhl': object
            })
            register_data.append(df)

    # Concatenate all the DataFrames into a single DataFrame
    table = pd.concat(register_data)

    table['name_last'] = table['name_last'].str.lower()
    table['name_first'] = table['name_first'].str.lower()
    table[['key_mlbam', 'key_fangraphs']] = table[['key_mlbam', 'key_fangraphs']].fillna(-1)
    table[['key_mlbam', 'key_fangraphs']] = table[['key_mlbam', 'key_fangraphs']].astype(int)
    return table


def player_id_lookup(last_name: str, first_name: str = None) -> pd.DataFrame:
    """
    Obtain IDs for single Player

    :param last_name: Last Name of Player
    :type last_name: string
    :param first_name: First Name of Player
    :type first_name: string
    """
    last_name = last_name.lower()
    if first_name:
        first_name = first_name.lower()
    table = get_lookup_table()
    if first_name is None:
        results = table.loc[table['name_last'] == last_name]
    else:
        results = table.loc[(table['name_last'] == last_name) & (table['name_first'] == first_name)]

    results = results.reset_index().drop('index', 1)
    return results


def player_id_reverse_lookup(player_ids, key_type=None) -> pd.DataFrame:
    """
    Retrieve a table of player information given a list of player ids

    :param List player_ids: list of player ids
    :param string key_type: name of the key player_type being looked up (one of "mlbam", "retro", "bbref", or "fangraphs")
    :return: DataFrame
    """
    key_types = ('mlbam', 'retro', 'bbref', 'fangraphs',)

    if not key_type:
        key_type = key_types[0]  # default is "mlbam" if key_type not provided
    elif key_type not in key_types:
        raise ValueError(
            '[Key Type: {}] Invalid; Key Type must be one of "{}"'.format(key_type, '", "'.join(key_types))
        )

    table = get_lookup_table()
    key = 'key_{}'.format(key_type)

    results = table[table[key].isin(player_ids)]
    results = results.reset_index().drop('index', 1)
    return results



import io
import logging
from datetime import datetime, timedelta, date
from typing import Literal

import numpy as np
import pandas as pd
import requests
from bs4 import BeautifulSoup

from plyball.utils import first_season_map


class BaseballReference(object):
    """
    Class for BaseballReference
    """

    urls = {
        'daily_war': 'http://www.baseball-reference.com/data/war_daily_{}.txt',
        'player_type': 'http://www.baseball-reference.com/leagues/daily.cgi?{}',
        'standings': 'http://www.baseball-reference.com/leagues/MLB/{}-standings.shtml',
        'team_results': 'http://www.baseball-reference.com/teams/{}/{}-schedule-scores.shtml'
    }

    def __get_league_stats_html(self, position_type: Literal['p', 'b'],
                                start_date: date,
                                end_date: date):
        """
        Get League Stats from Baseball Reference for pitching and batting.

        :param DateTime start_date: Date to Begin Stats Lookup
        :param DateTime end_date: Date to End Stats Lookup
        :param string position_type: pitching (p) or batting (b)
        :return: BeautifulSoup
        """

        parameters = {
            'user_team': '',
            'bust_cache': '',
            'type': position_type,
            'lastndays': '7',
            'dates': 'fromandto',
            'fromandto': '{}.{}'.format(start_date.strftime("%Y-%m-%d"), end_date.strftime("%Y-%m-%d")),
            'level': 'mlb',
            'franch': '',
            'stat': '',
            'stat_value': '0',
        }

        logging.info(self.urls['player_type'].format('&'.join(['{}={}'.format(k, v)
                                                               for k, v in parameters.items()])))
        s = requests.get(self.urls['player_type'].format('&'.join(['{}={}'.format(k, v)
                                                                   for k, v in parameters.items()]))).content

        return BeautifulSoup(s, "lxml")

    def __get_league_stats_table(self,
                                 position_type: Literal['b', 'p'],
                                 start_date: date = datetime.today(),
                                 end_date: date = datetime.today() - timedelta(1)):
        """
        Get player_type from Baseball Reference for a certain time frame.

        :param string position_type: Pitching (p) or Batting (b)
        :param DateTime start_date: Beginning Date
        :param DateTime end_date: Ending Date
        :return: DataFrame
        """
        soup = self.__get_league_stats_html(position_type, start_date, end_date)
        table = soup.find_all('table')[0]
        data = []
        headings = [th.get_text() for th in table.find("tr").find_all("th")][1:]
        data.append(headings)
        table_body = table.find('tbody')
        rows = table_body.find_all('tr')
        for row in rows:
            cols = row.find_all('td')
            cols = [ele.text.strip() for ele in cols]
            data.append([ele for ele in cols])
        data = pd.DataFrame(data[1:], columns=data[0])
        return data

    def __get_team_result_html(self, season: int, team: str):
        s = requests.get(self.urls['team_results'].format(team, season)).content
        return BeautifulSoup(s, "lxml")

    @staticmethod
    def __process_win_streak(data):
        if data['Streak'].count() > 0:
            data['Streak2'] = data['Streak'].str.len()
            data.loc[data['Streak'].str[0] == '-', 'Streak2'] = -data['Streak2']
            data['Streak'] = data['Streak2']
            data = data.drop('Streak2', 1)
        return data

    def schedule_and_record(self, season: int, team: str):
        """
        Get a teams schedule and record from Baseball Reference

        :param season: Season Year
        :param team: Team Initials
        :return: DataFrame
        """
        try:
            if season < first_season_map[team]:
                m = "Season cannot be before first_name year of a team's existence"
                raise ValueError(m)
        # ignore validation if team isn't found in dictionary
        except KeyError:
            pass
        if season > datetime.now().year:
            raise ValueError('Season cannot be after current year')

        team_results = self.__get_team_results_table(season, team)
        team_results = self.__process_win_streak(team_results)
        for column in ["R", "RA", "Inn", "Rank", "Attendance"]:
            team_results[column] = pd.to_numeric(team_results[column], errors='coerce')
        return team_results

    def __get_team_results_table(self, season: int, team: str):
        try:
            table = self.__get_team_result_html(season, team).find_all('table')[0]
        except:
            raise ValueError("Data cannot be retrieved for this team/year combo. Please verify that your team "
                             "abbreviation is accurate and that the team existed during the season you are searching "
                             "for.")
        data = []
        headings = [th.get_text() for th in table.find("tr").find_all("th")]
        headings = headings[1:]  # the "gm#" heading doesn't have a <td> element
        headings[3] = "Home_Away"
        data.append(headings)
        table_body = table.find('tbody')
        rows = table_body.find_all('tr')
        for row_index in range(len(rows) - 1):  # last_name row is a description of column meanings
            row = rows[row_index]
            try:
                cols = row.find_all('td')
                # links = row.find_all('a')
                if cols[1].text == "":
                    cols[
                        1].string = team
                if cols[3].text == "":
                    cols[3].string = 'Home'
                if cols[12].text == "":
                    cols[12].string = "None"
                if cols[13].text == "":
                    cols[13].string = "None"
                if cols[14].text == "":
                    cols[14].string = "None"
                if cols[8].text == "":
                    cols[8].string = "9"
                if cols[16].text == "":
                    cols[16].string = "Unknown"
                if cols[15].text == "":
                    cols[15].string = "Unknown"
                if cols[17].text == "":
                    cols[17].string = "Unknown"

                cols = [ele.text.strip() for ele in cols]
                data.append([ele for ele in cols if ele])
            except:
                # two cases will break the above: games that haven't happened yet, and BR's redundant mid-table headers
                # if future games, grab the scheduling info. Otherwise do nothing.
                if len(cols) > 1:
                    cols = [ele.text.strip() for ele in cols][0:5]
                    data.append([ele for ele in cols if ele])
        data = pd.DataFrame(data, columns=data.iloc[0])
        data['Attendance'].replace(r'^Unknown$', np.nan, regex=True, inplace=True)
        return data

    def get_stats_range(self, position_type: Literal['p', 'b'], start_dt: date = None, end_dt: date = None):
        """
        Get player_type for a set time range. This can be the past week, the
        month of August, anything. Just supply the start and end date in YYYY-MM-DD
        format.

        :param string position_type: Pitching (p) or Batting (b)
        :param DateTime start_dt:
        :param DateTime end_dt:
        :return: DataFrame
        """
        """

        """
        table = self.__get_league_stats_table(position_type, start_dt, end_dt)

        if position_type == 'p':
            table = table.replace('---%', np.NaN)
            for column in ['Age', '#days', 'G', 'GS', 'W', 'L', 'SV', 'IP', 'H',
                           'R', 'ER', 'BB', 'SO', 'HR', 'HBP', 'ERA', 'AB', '2B',
                           '3B', 'IBB', 'GDP', 'SF', 'SB', 'CS', 'PO', 'BF', 'Pit',
                           'WHIP', 'BAbip', 'SO9', 'SO/W']:
                table[column] = pd.to_numeric(table[column])
            # convert str(xx%) values to float(0.XX) decimal values
            for column in ['Str', 'StL', 'StS', 'GB/FB', 'LD', 'PU']:
                table[column] = pd.to_numeric(table[column].replace('%', '', regex=True), errors='coerce') / 100
        elif position_type == 'b':
            for column in ['Age', '#days', 'G', 'PA', 'AB', 'R', 'H', '2B', '3B',
                           'HR', 'RBI', 'BB', 'IBB', 'SO', 'HBP', 'SH', 'SF', 'GDP',
                           'SB', 'CS', 'BA', 'OBP', 'SLG', 'OPS']:
                table[column] = pd.to_numeric(table[column], errors='coerce')
        table = table.drop('', 1)
        return table

    def get_season_stats(self, position_type: Literal['p', 'b'], season: int = datetime.now().year):
        """
        Get Stats from a set season.
        :param position_type: Pitching (p) or Batting (b)
        :param season: Season Year
        :return: DataFrame
        """

        start_dt = datetime(season, 3, 1)
        end_dt = datetime(season, 11, 1)
        return self.get_stats_range(position_type, start_dt, end_dt)

    def get_daily_war(self, position_type: Literal['p', 'b']):
        """
        Get data from Daily War tables (pitching or batting). Returns WAR, its components, and a few other useful player_type.

        :param string position_type: pitch or bat
        :return: DataFrame
        """
        s = requests.get(self.urls['daily_war'].format(position_type)).content
        c = pd.read_csv(io.StringIO(s.decode('utf-8')))
        return c
